/*
 * W2000A Screenshot Utility v0.3
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifndef WIN32
#include <err.h>
#else
#include "ComTools.h"
#endif

#define MAXPLANES	16

/*
 * XXX: Please note:
 *      This is _not_ 100% identical to the actual
 *      output seen on the DSO.  For example:
 *      The math channel is drawn _over_
 *      the browse bar.  Put the Math channel
 *      behind UI in the planeorder to witness
 *      the actual output.
 */
static const int planeorder[] = {
	1,			/* Grid */
	7, 8, 9, 10, 		/* Channels 1-4 */
				/* XXX: Red Arrow and Stop-Sign in UI on 10/Channel4? */
	11,			/* Math Channel */
	15, 16,			/* Cursors and Trigger Mark */
	4, 5, 6, 2, 3,		/* UI */
	12, 13, 14,		/* Not yet determined */
	0
};

static const int planecolours[][3] = {
			/* R     G     B */
/* Grid_Plane */	{ 0x3F, 0x3F, 0x3F }, /* dark gray */
/* UI_Plane1 */		{ 0xFB, 0xFB, 0xFB }, /* dark white */
/* UI_Plane2 */		{ 0x00, 0x00, 0x00 }, /* black */
/* UI_Plane3 */		{ 0xFF, 0xFD, 0xEE }, /* yellow tinted white */
/* UI_Plane4 */		{ 0xE6, 0xE6, 0x83 }, /* beige-ish yellow */
/* UI_Plane5 */		{ 0x7E, 0x7E, 0x7E }, /* gray */
/* Channel_Plane1 */	{ 0xFF, 0xFF, 0x00 }, /* yellow */
/* Channel_Plane2 */	{ 0x90, 0xEE, 0x90 }, /* light green */
/* Channel_Plane3 */	{ 0x32, 0xAB, 0xFF }, /* light blue */
/* Channel_Plane4 */	{ 0xFF, 0x00, 0x00 }, /* red */
/* Channel_Math_Plane */{ 0xFF, 0x96, 0xF6 }, /* pink */
/* Memory_Plane1 */	{ 0x00, 0x00, 0x00 }, /* n.y.d. */
/* Memory_Plane2 */	{ 0x00, 0x00, 0x00 }, /* n.y.d. */
/* Memory_Plane3 */	{ 0x00, 0x00, 0x00 }, /* n.y.d. */
/* Marker_Plane1 */	{ 0xFF, 0x67, 0x00 }, /* orange-red */
/* Marker_Plane2 */	{ 0xFF, 0x67, 0x00 }, /* orange-red */
			{ 0x00, 0x00, 0x00 }  /* don't remove this! */
};

static unsigned char pixmaps[MAXPLANES][640*480];
static char fnamprefix[256];
static int rdplanes;
#ifdef WIN32
static int com;
#endif

#ifdef WIN32
void err(int e, const char *fmt, ...)
{

	/*
	 * Don't even know if win32 has sth. like vsnprintf(),
	 * so we don't care.
	 */
	(void)fprintf(stderr, "Error: %s, exiting.\n", fmt);
	exit(e);
}
#endif

static void
usage(void)
{

	(void)fprintf(stderr,
			"Usage:\n"
#ifdef WIN32
			"w2000a-screenshot.exe [-h] [-c n] [-f prefix] [-b]\n"
#else
			"w2000a-screenshot [-h] [-f prefix] [-b]\n"
#endif
			" -h\t\tthis help\n"
#ifdef WIN32
			" -c n\t\tCOM port (0 < n < 10)\n"
#endif
			" -f prefix\toutput file\n"
			" -b write as BMP file, not P(P|G)M\n"
			"\n"
#ifdef WIN32
			"Input is read from specified COM port,\n"
			"the default is COM1.\n"
#else
			"Input is read from stdin, diagnostics and\n"
			"status messages are directed to stderr.\n"
#endif
			"\n"
			"If P(P|G)M (default) output is in use, planes\n"
			"are written to \"[prefix]_dd.pgm\", with\n"
			"\"dd\" as the two digit plane sequence number.\n"
			"Prefix defaults to \"screenshot\".\n"
			"\n"
			"Planes are coloured and combined in \"[prefix].[suf]\",\n"
			"with suffix ppm or bmp.\n");
	exit(1);
}

static int
rdchar(void)
{
#ifdef WIN32
	while (ComGetReadCount(com) < 1)
		(int)usleep(1000);
	return ComRead(com);
#else
	unsigned char c;

	if (read(0, &c, 1) > 0)
		return c;
	return -1;
#endif
}

static void
retrieve(int all)
{
	FILE *fp;
	char fnam[256];
	int r, pnum, rd, total, ppos;
	unsigned char c, l, b;
	int i;

#ifdef WIN32
	if (!ComOpen(com, 115200, P_NONE, S_1BIT, D_8BIT))
		err(1, "Unable to open COM port");
#endif
	(void)fprintf(stderr, "* Waiting for Screenshot...\n");

	l = 0;
	while ((r = rdchar()) > 0) {
		c = (unsigned char)r;
		if (c == 255 && l == 'S')
			break;
		l = c;
	}
	if (r <= 0)
		exit(1);

	(void)fprintf(stderr, "* Found Screenshot Start Marker\n");

	total = 0;
	rdplanes = 0;
	fp = NULL;
	for (pnum = 0; pnum < MAXPLANES; pnum++) {
		(void)fprintf(stderr, "  - Receiving Plane #%.2d...\n", pnum+1);

		if (all) {
			(void)snprintf(fnam, sizeof(fnam), "%s_%.2d.pgm", fnamprefix, pnum+1);
			if ((fp = fopen(fnam, "w+")) == NULL)
				err(1, "fopen(%s, w+) failed", fnam);

			(void)fprintf(fp, "P5\n640 480\n255 ");
		}

		rd = 0;
		ppos = 0;
		while ((r = rdchar()) > 0) {
			c = (unsigned char)r;
			rd++;
			if (rd % 100 == 0)
				(void)fprintf(stderr, "\r    %d bytes...", rd);
			if (c == 255)
				break;
			b = 0x00;
			if (c & (1 << 7))
				b = 0xff;
			c &= ~(1 << 7);

			for (i = 0; i < c; i++) {
				pixmaps[pnum][ppos++] = b;
				if (all)
					(void)fputc(b, fp);
			}
		}
		total += rd;
		rdplanes++;

		(void)fprintf(stderr, "\r    Read %d bytes.\n", rd);

		if (all && fclose(fp) != 0)
			err(1, "fclose(%s) failed", fnam);

		if ((r = rdchar()) > 0) {
			c = (unsigned char)r;
			if (c == 'E') {
				(void)fprintf(stderr, "* End of Transmission\n");
				break;
			} else if (c == 0xFF)
				(void)fprintf(stderr, "    (another plane follows)\n");
		} else {
			(void)fprintf(stderr, "* Short read, ignoring\n");
			break;
		}
	}

	(void)fprintf(stderr, "* Total bytes transferred: %d\n", total);
}

static void
combine(int bmp)
{
	FILE *fp;
	char fnam[256], suf[4];
	int ppos, i, plane, colour;

	(void)fprintf(stderr, "* Combining...\n");

	(void)snprintf(suf, sizeof(suf), "ppm");
	if (bmp)
		(void)snprintf(suf, sizeof(suf), "bmp");

	(void)snprintf(fnam, sizeof(fnam), "%s.%s", fnamprefix, suf);
	if ((fp = fopen(fnam, "w+")) == NULL)
		err(1, "fopen(%s, w+) failed", fnam);

	if (bmp) {
		/*
		 * http://en.wikipedia.org/wiki/BMP_file_format
		 */
		/* 0x00 */ (void)fprintf(fp, "BM");
		/* 0x02 */ (void)fprintf(fp, "%c%c%c%c%c%c%c%c", 0, 0, 0, 0, 0, 0, 0, 0);
		/* 0x0A */ (void)fprintf(fp, "%c%c%c%c", 54, 0, 0, 0);		/* 40 (header size) + 0x0A + 4 */
		/* 0x0E */ (void)fprintf(fp, "%c%c%c%c", 40, 0, 0, 0);		/* header size */
		/* 0x12 */ (void)fprintf(fp, "%c%c%c%c", 128, 2, 0, 0);		/* 640px */
		/* 0x16 */ (void)fprintf(fp, "%c%c%c%c", 32, 254, 255, 255);	/* -480px */
		/* 0x1A */ (void)fprintf(fp, "%c%c", 1, 0);
		/* 0x1C */ (void)fprintf(fp, "%c%c", 24, 0);			/* bits per pixel */
		/* 0x1E */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);		/* no compression */
		/* 0x22 */ (void)fprintf(fp, "%c%c%c%c", 0, 16, 14, 0);		/* bitmap data size: 640*480*3=921600 Bytes */
		/* 0x26 */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);
		/* 0x2A */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);
		/* 0x2E */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);
		/* 0x32 */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);
	} else {
		/*
		 * http://netpbm.sourceforge.net/doc/ppm.html
		 */
		(void)fprintf(fp, "P6\n640 480\n255 ");
	}

	for (ppos = 0; ppos < 640*480; ppos++) {
		colour = MAXPLANES; /* default colour, black */
		for (i = 0; i < MAXPLANES && i < rdplanes; i++) {
			plane = planeorder[i] - 1;
			if (pixmaps[plane][ppos] == 0xFF)
				colour = plane;
		}
		if (bmp) {
			/* colour order BGR */
			(void)fputc(planecolours[colour][2], fp);
			(void)fputc(planecolours[colour][1], fp);
			(void)fputc(planecolours[colour][0], fp);
		} else {
			/* colour order RGB */
			(void)fputc(planecolours[colour][0], fp);
			(void)fputc(planecolours[colour][1], fp);
			(void)fputc(planecolours[colour][2], fp);
		}
	}

	if (fclose(fp) != 0)
		err(1, "fclose(%s) failed", fnam);

	(void)fprintf(stderr, "* Done\n");
}

int
main(int argc, char **argv)
{
	int ch, bflag;

#ifdef WIN32
	com = 0;
#endif
	bflag = 0;
	(void)snprintf(fnamprefix, sizeof(fnamprefix), "screenshot");
#ifdef WIN32
	while ((ch = getopt(argc, argv, "c:f:bh")) != -1) {
#else
	while ((ch = getopt(argc, argv, "f:bh")) != -1) {
#endif
		switch (ch) {
#ifdef WIN32
		case 'c':
			com = atoi(optarg) - 1;
			if (com < 0 || com > 8)
				err(1, "invalid COM port given, use -c n, where 0 < n < 10");
			break;
#endif
		case 'f':
			(void)snprintf(fnamprefix, sizeof(fnamprefix), "%s", optarg);
			break;
		case 'b':
			bflag = 1;
			break;
		case 'h':
		default:
			usage();
			/* NOTREACHED */
		}
	}

	retrieve(!bflag);
	combine(bflag);

	return 0;
}
